import logging


def image_edge_definer(pixel_positions, boundaries, pixels_to_edge):
    # Mosaic image position coordinates are from the centre of each
    # image, according to Cockpit's absolute position
    position_on_mosaic = [pixel_positions[0] - boundaries[0, 0],
                          pixel_positions[1] - boundaries[0, 1]]
    start = (position_on_mosaic[0] - pixels_to_edge[0],
             position_on_mosaic[1] - pixels_to_edge[1])
    end = (position_on_mosaic[0] + pixels_to_edge[0],
           position_on_mosaic[1] + pixels_to_edge[1])
    logging.debug("Image edges defined as (%i, %i), (%i, %i)", *start, *end)
    return start, end


def marker_edge_definer(pixel_positions, boundaries, pixels_to_edge):
    # Markers position coordinates are from the top right corner
    # of each image, according to Cockpit's absolute position
    # NOTE: "top right" is according to the final display
    # directions NOT the pre-processed coordinates
    im_y_size = boundaries[1, 1] - boundaries[0, 1]
    position_on_mosaic = [pixel_positions[0] - boundaries[0, 0],
                          pixel_positions[1] - boundaries[0, 1]]
    start = (position_on_mosaic[0] - pixels_to_edge[0] * 2,
             im_y_size - (position_on_mosaic[1]))
    end = (position_on_mosaic[0],
           im_y_size - (position_on_mosaic[1] - pixels_to_edge[1] * 2))
    logging.debug("Marker edges defined as (%i, %i), (%i, %i)", *start, *end)
    return start, end



version_info = (1, 5, 0)
__version__ = '.'.join(str(c) for c in version_info)
__author__ = "Thomas Fish"

def stitch_and_save(mosaic_file, marker_file=None):
    """
    PARAMETERS:
        mosaic_file - Path of .txt file that contains the mosaic information, including the path to the .mrc file
        marker_file - Path of .txt file that contains a list of marker placements and associated numbers (please make sure this correctly corresponds to the mosaic file)
    
    The output will be saved as the mosaic filename, with the suffix '.ome.tiff' (or '_marked.ome.tiff' if markers are supplied), in same directory as the mosaic file.
    """
    from .run import main_run
    from .file_handler import get_config

    config, _ = get_config()
    main_run(config, mosaic_file, marker_file)


def stitch(mosaic_file, marker_file=None):
    """
    PARAMETERS:
        mosaic_file - Path of .txt file that contains the mosaic information, including the path to the .mrc file
        marker_file - Path of .txt file that contains a list of marker placements and associated numbers (please make sure this correctly corresponds to the mosaic file)
    
    OUTPUT:
        mosaic (numpy array)
        metadata (as an editable omexml object)
        tiff_file (the default path stitch_and_save would save to, as a string)
    """
    from .run import _stitch
    from .file_handler import get_config
    
    config, _ = get_config()
    return _stitch(config, mosaic_file, marker_file)


def save(mosaic, metadata, tiff_file):
    """
    PARAMETERS:
        mosaic - numpy array
        metadata - omexml object from the stitch function (or omexml-dls)
        tiff_file - Path which the mosaic should be saved to (as a string)
    
    The output will be saved as the mosaic filename, with the suffix '.ome.tiff' (or '_marked.ome.tiff' if markers are supplied), in same directory as the mosaic file.
    """
    from .run import _save

    _save(mosaic, metadata, tiff_file)



import numpy as np
import logging
from numpy.testing import assert_array_equal

from .edge_definer import image_edge_definer
from .image_normaliser import normalise_to_datatype, cast_to_dtype


class Stitcher():

    def __init__(self, datatype="uint16"):
        '''Default datatype="uint16"'''
        self.dtype = np.dtype(datatype)
        self.brightfield_list = []

    def make_mosaic(self, unstitched, fl_filter=True, normalise=True):
        logging.info("Creating mosaic")
        if unstitched.img_count == unstitched.images.shape[0]:
            if fl_filter:
                self.brightfield_list = self._find_brightfield_images(unstitched.img_count, unstitched.exposure_minmax)
            else:
                self.brightfield_list = [i for i in range(unstitched.img_count)]
            # create new large array and load data into it from mosaic:
            mosaic_size = (unstitched.boundaries[1, 0] - unstitched.boundaries[0, 0],
                         unstitched.boundaries[1, 1] - unstitched.boundaries[0, 1])
            
            # If we are not normalising, fill with zero values rather than max
            fill_value = np.iinfo(self.dtype).max * normalise
            mosaic_array = np.full(mosaic_size, fill_value, dtype=self.dtype)
            
            images = unstitched.images[self.brightfield_list]  # Filter out unwanted images
            unstitched.clear_image_array()
            if normalise:
                # Rescale max/min to fit data type
                images = normalise_to_datatype(images, self.dtype, trim=True)
            
            # Cast to output data type
            images = cast_to_dtype(images, self.dtype)

            for i in range(len(self.brightfield_list)):
                start, end = image_edge_definer(
                    unstitched.pix_positionlist[self.brightfield_list[i], :],
                    unstitched.boundaries,
                    unstitched.pix2edge
                    )
                # Array needs to be transposed for python versus dv.
                # This rotates each image so they line up correctly
                mosaic_array[start[0]:end[0], start[1]:end[1]] = images[i, :, :].T
            del(images)
            # Rotate back and flip
            return np.flip(mosaic_array.T, 0)
        else:
            logging.error("Number of images doesn't match between files")
            raise AssertionError("Number of images doesn't match between files")

    def get_brightfield_list(self):
        return self.brightfield_list

    @staticmethod
    def _find_brightfield_images(img_count, minmax):
        """
        This returns a list of indices of "good" images (i.e. not fluorescent 
        images).
        This is based on the min values of fluorescent images being 
        lower than the (median - std deviation) of the min of all the  
        images.

        This method is not valid if there are too many fl images to bf images,
        as this will affect the median value, leading to few, if any, images 
        being filtered.

        If more than half of the images are filtered out by this metric,
        it is assumed something hasn't worked correctly and a list including
        of all image indices will be returned.
        """
        full_list = range(0, img_count)

        median_min = np.median(minmax[:, 0])
        std_min = np.std(minmax[:, 0])

        good_list = np.nonzero(
            minmax[:, 0] > median_min - std_min
            )[0].tolist()

        if len(good_list) == img_count:
            logging.info("No potential fluorescence images found.")
        else:
            logging.info(
                "Potential fluorescence images found: '%s' (counted from 0)",
                ", ".join(map(str, np.setdiff1d(full_list, good_list)))
                 )
        if len(good_list) > img_count / 2:
            return good_list
        return full_list



import os
import sys
import unittest
from unittest.mock import patch, MagicMock, ANY
from subprocess import Popen, PIPE, STDOUT
from time import sleep

from stitch_m.__init__ import __version__ as version

from stitch_m.scripts import commandline

class TestEntryPointCommandline(unittest.TestCase):
# Before these running tests, the module must be installed in development mode using:
# `pip install --editable . --user` (from the project folder)
# it can be uninstalled at the end using:
# `pip uninstall StitchM`
# IF TESTING ON WINDOWS: make sure the correct python installation is set as the default in PATH

    @classmethod
    def setUpClass(cls):
        # Set maximum difference string length to None (infinite)
        cls.maxDiff = None
        # Add local intall to path for testing with development install
        if os.name == "posix":
            os.environ["PATH"] += os.pathsep + os.path.join(os.path.expanduser("~"), ".local", "bin")
    
    @patch('argparse.ArgumentParser.print_help')
    def test_commandline_function(self, mocked_printhelp):
        sys.argv = ["StitchM"]
        commandline.main()

        mocked_printhelp.assert_called_once()

    @patch('argparse.ArgumentParser.print_help')
    def test_commandline_function_setup_subpaser(self, mocked_printhelp):
        sys.argv = ["StitchM", "setup"]
        commandline.main()

        mocked_printhelp.assert_called_once()

    def test_commandline_function_setup_config(self):
        with patch('stitch_m.file_handler.create_user_config', MagicMock()) as mocked_config_creator:
            sys.argv = ["StitchM", "setup", "--config"]
            commandline.main()

            mocked_config_creator.assert_called_once()

    def test_commandline_function_setup_windows_shortcut(self):
        with patch('stitch_m.file_handler.create_Windows_shortcut', MagicMock()) as mocked_shortcut_creator:
            sys.argv = ["StitchM", "setup", "-w"]
            commandline.main()

            mocked_shortcut_creator.assert_called_once()

    def test_commandline_function_with_args(self, ):
        with patch('stitch_m.run.main_run', MagicMock()) as main_run:
            kwargs = {}
            kwargs['mosaic'] = "mosaic_path"
            kwargs['markers'] = "marker_path"
            kwargs['normalise'] = True
            kwargs['fl_filter'] = True

            sys.argv =["StitchM", "--mosaic", kwargs['mosaic'], "--markers", kwargs['markers']]
            commandline.main()

            main_run.assert_called_once_with(ANY, **kwargs)

    def test_commandline_function_with_args_normalise_filter(self, ):
        with patch('stitch_m.run.main_run', MagicMock()) as main_run:
            kwargs = {}
            kwargs['mosaic'] = "mosaic_path"
            kwargs['markers'] = "marker_path"
            kwargs['normalise'] = False
            kwargs['fl_filter'] = False

            sys.argv =["StitchM", "--mosaic", kwargs['mosaic'], "--markers", kwargs['markers'], "--unfiltered", "--no-normalisation"]
            commandline.main()

            main_run.assert_called_once_with(ANY, **kwargs)

    def test_commandline_method(self):
        args = ["path_to/mosaic"]
        run_args = ["StitchM", "--mosaic", args[0]]
        with Popen(run_args, universal_newlines=True, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as p:
            sleep(2)
            stdout, _ = p.communicate("pressed key", timeout=5)
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(f"Invalid arguments: {args[0]}, None, True", stdout, msg=f"Actual stdout: {stdout}")

    def test_commandline_method_invalid_path(self):
        args = ["path_to/mosaic.txt"]
        run_args = ["StitchM", "--mosaic", args[0]]
        with Popen(run_args, universal_newlines=True, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as p:
            sleep(2)
            stdout, _ = p.communicate("pressed key", timeout=5)
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(f"Mosaic file path cannot be resolved", stdout, msg=f"Actual stdout: {stdout}")

    def test_commandline_method_with_marker_file(self):
        args = ["path_to/mosaic", "path_to/markers"]
        run_args = ["StitchM", "--mosaic", args[0], "--markers", args[1], "--no-normalisation"]
        with Popen(run_args, universal_newlines=True, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as p:
            sleep(2)
            stdout, _ = p.communicate("pressed key", timeout=5)
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(f"Invalid arguments: {args[0]}, {args[1]}, False", stdout, msg=f"Actual stdout: {stdout}")

    def test_commandline_method_version_number(self):
        run_args = ["StitchM", "--version"]
        with Popen(run_args, universal_newlines=True, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as p:
            sleep(2)
            stdout, _ = p.communicate("pressed key", timeout=5)
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertEqual(f"StitchM {version}", stdout, msg=f"Actual stdout: {stdout}")

    def test_commandline_method_setup_subparser(self):
        run_args = ["StitchM", "setup"]
        with Popen(run_args, universal_newlines=True, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as p:
            sleep(2)
            stdout, _ = p.communicate("pressed key", timeout=5)
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(f"StitchM setup [-w] [-c] [-h]", stdout, msg=f"Actual stdout: {stdout}")



"""
For running using Scalene (https://github.com/plasma-umass/scalene)

Run with:
`scalene src/tests/scalene.py` for just the overall stats
move scalene.py to the top level and run with `scalene scalene.py` for finer detail
"""
 
from os import path, remove
from pathlib import Path

from stitch_m import stitch_and_save

base_path = Path("/dls/science/groups/das/ExampleData/B24_test_data/StitchM_test_data/files")
test_files = (
    base_path / "B15Grid2.txt",
    base_path / "B8G1-IR_mosaic.txt",
    base_path / "B8G2-IR_mosaic.txt",
    base_path / "Fid_T2G3_mosaic.txt",
    base_path / "Yo10_G3_mosaic.txt",
    base_path / "PP4_G2_mosaic.txt"
    )

test_marker_files = (
    base_path / "B15_location_markers.txt",
    base_path / "B8G1-IR_markers.txt",
    base_path / "B8G2-IR_markers.txt",
    base_path / "Fid_T2G3_markers.txt",
    base_path / "Yo10_G3_mosaic_MARKERS.txt",
    base_path / "PP4_G2_markers.txt"
    )

for mos, mar in zip(test_files, test_marker_files):
    stitch_and_save(mos, mar)
    output_path = str(mos).replace('.txt', '_marked.ome.tiff')

    if path.isfile(output_path):
        remove(output_path)



import os
import sys
from pathlib import Path
import logging

level_dict = {
    "1": logging.DEBUG,
    "2": logging.INFO,
    "3": logging.WARNING,
    "4": logging.ERROR,
    "5": logging.CRITICAL,
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warn": logging.WARNING,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
    }

class LogHandler():

    def __init__(self, file_level="debug", stream_level="info", log_to_file=False, backup_count=10, config=None, config_messages=None):
        self.logger = self._create_or_setup(file_level, stream_level, log_to_file, backup_count, config, config_messages)

    def __enter__(self, file_level="debug", stream_level="info", log_to_file=False, backup_count=10, config=None, config_messages=None):
        return self.logger

    def __exit__(self, et, ev, tb):
        self.close()

    def get_logger(self):
        return self.logger

    def close(self):
        handlers = self.logger.handlers
        for handler in handlers:
            handler.close()
            self.logger.removeHandler(handler)
        logging.shutdown()

    def _create_or_setup(self, file_level="debug", stream_level="info", log_to_file=False, backup_count=10, config=None, config_messages=None):
        if config:
            logger = self.setup_logging(config)
        else:
            logger = self.create_logger(file_level, stream_level, log_to_file, backup_count)
        
        if config_messages:
            for msg in config_messages:
                logging.warning(msg)
            
        return logger

    def setup_logging(self, config):
        from configparser import Error as ConfigError
        from .file_handler import boolean_config_handler
        try:
            return self.create_logger(
                config.get('LOGGING', 'file level'),
                config.get('LOGGING', 'stream level'),
                boolean_config_handler(config, 'LOGGING', 'log to file', default=False),
                backup_count=10
                )
        except ConfigError:
            # Create logger with defaults and log issue
            logging.error("Error retrieving logging levels from config file. Using verbose settings.")
            return self.create_logger()
            

    @staticmethod
    def create_logger(file_level="debug", stream_level="info", log_to_file=False, backup_count=10):
        """
        Creates a logging file handler that creates a new file daily at midnight, with up to 'backup_count' log files saved.
        This also sets logging for stdout to info level.
        
        INPUTS:
        file_level (string or int), stream_level (string or int), backup_count (int)
        
        ACCEPTABLE LOGGING LEVELS:
        debug OR 1
        info OR 2
        warn OR warning OR 3
        error OR 4
        Critical OR 5
        
        Alternatively will default to debug.
        """
        if stream_level in level_dict:
            stream_level = level_dict[stream_level]
        else:
            # use debug if no valid logging level is passed
            stream_level = logging.ERROR
            logging.error("Invalid logging level passed. Stream logging will be set to ERROR")

        stream_handler = logging.StreamHandler(sys.stdout)
        stream_format = logging.Formatter("%(levelname)-8s %(message)s")
        stream_handler.setFormatter(stream_format)

        logger = logging.getLogger()
        logger.setLevel(logging.NOTSET)
        stream_handler.setLevel(stream_level)
        logger.addHandler(stream_handler)

        if log_to_file:
            from stitch_m.file_handler import get_user_log_path
            from logging.handlers import TimedRotatingFileHandler
            logpath = get_user_log_path()
            if logpath:
                if file_level in level_dict:
                    file_level = level_dict[file_level]
                else:
                    # use debug if no valid logging level is passed
                    file_level = logging.DEBUG
                    logging.error("Invalid logging level passed. File logging will be set to DEBUG")
                file_handler = TimedRotatingFileHandler(str(logpath / "stitchm_log"), when='midnight', interval=1, backupCount=backup_count)
                file_format = logging.Formatter("%(asctime)s %(levelname)-8s %(filename)s|%(funcName)s: %(message)s", datefmt="%d %b %Y - %H:%M:%S")
                file_handler.setFormatter(file_format)
                file_handler.setLevel(file_level)
                logger.addHandler(file_handler)
        return logger



import unittest
from unittest.mock import patch, MagicMock, ANY

from stitch_m import stitch_and_save


class TestPackageFunctions(unittest.TestCase):
    def test_module_stitch_and_save_method(self):
        with patch('stitch_m.run.main_run', MagicMock()) as stitch_m_run:
            args = ["path_to/mosaic.txt", None]
            stitch_and_save(args[0])
            stitch_m_run.assert_called_once_with(ANY, *args)

    def test_module_stitch_and_save_method_with_marker_file(self):
        with patch('stitch_m.run.main_run', MagicMock()) as stitch_m_run:
            args = ["path_to/mosaic.txt", "path_to/markers.txt"]
            stitch_and_save(*args)
            stitch_m_run.assert_called_once_with(ANY, *args)




import os
import sys
import logging
import argparse

def main():
    from .__init__ import __version__
    from .file_handler import get_user_config_path

    description = "Stitch mosaics from Cockpit with (or without) ROIs."
    config_path, _ = get_user_config_path()
    if config_path is not None and config_path.exists():
        description += f" User config can be found at {config_path}."

    parser = argparse.ArgumentParser(prog="StitchM", description=description, add_help=False)
    stitch_group = parser.add_argument_group("Stitching arguments")
    stitch_group.add_argument(
        "-m", "--mosaic", metavar="PATH/TO/MOSAIC_FILE.TXT", dest="mosaic",
        type=str, default="", action='store',
        help="the mosaic to be stitched (.txt file)")
    stitch_group.add_argument(
        "-a", "--markers", metavar="PATH/TO/MARKER_FILE.TXT", dest="markers",
        type=str, default="", action='store',
        help="[OPTIONAL] the markers to be added as ROIs (.txt file)")
    stitch_group.add_argument(
        "-n", "--no-normalisation", dest="normalise",
        default=True, action='store_false',
        help="do not normalise images")
    stitch_group.add_argument(
        "-u", "--unfiltered", dest="fl_filter",
        default=True, action='store_false',
        help="do not try to filter out fluorescence images")

    setup_subparsers = parser.add_subparsers(title="Setup options", description="enter `StitchM setup -h` for details")
    setup_parser = setup_subparsers.add_parser(name="setup", add_help=False)
    setup_parser.add_argument(
        "-w", "--windows-shortcut", dest="windows_shortcut",
        action='store_true',
        help=(
            "creates a Windows shortcut on the user's desktop that accepts drag " +
            "and dropped files (one mosaic at a time, optionally including markers)"))
    setup_parser.add_argument(
        "-c", "--config", dest="config",
        action='store_true',
        help="creates a user specific config if called")
    
    setup_info_group = setup_parser.add_argument_group("Setup info")
    setup_info_group.add_argument('-h', '--help', action='help', help="show this help message and exit")

    package_group = parser.add_argument_group("Package info")
    package_group.add_argument('-v', '--version', action='version', version="%(prog)s {}".format(__version__))
    package_group.add_argument('-h', '--help', action='help', help="show this help message and exit")

    # if args has the attribute 'config', the setup subparser has been called and args.win will also exist
    if "setup" in sys.argv:
        from .log_handler import LogHandler
        with LogHandler("info", "info"):
            setup_args = parser.parse_args(namespace=setup_parser)
            if setup_args.windows_shortcut:
                from .file_handler import create_Windows_shortcut
                create_Windows_shortcut()
            if setup_args.config:
                from .file_handler import create_user_config
                create_user_config()
            if not setup_args.windows_shortcut and not setup_args.config:
                setup_parser.print_help()
        return

    args = parser.parse_args()
    # Empty strings are False
    if args.mosaic:
        from .file_handler import get_config
        from .log_handler import LogHandler
        from .run import main_run

        try:
            config, config_messages = get_config()
            with LogHandler(config=config, config_messages=config_messages):
                kwargs = vars(args)
                main_run(config, **kwargs)
        except Exception:
            logging.error('Error has occurred. Please see traceback for more info.', exc_info=True)
    else:
        parser.print_help()
